namespace Commonmark.Internal {

    public class HtmlBlockParser : Parser.Block.AbstractBlockParser {

        private static GLib.Regex[,] BLOCK_PATTERNS;

        private Node.HtmlBlock block;
        private GLib.Regex closing_pattern;

        private bool finished;
        private BlockContent content;

        private HtmlBlockParser(GLib.Regex closing_pattern) {
            block = new Node.HtmlBlock();
            content = new Internal.BlockContent();
            this.closing_pattern = closing_pattern;
        }

        public override Node.Block get_block() {
            return block;
        }

        public override Parser.Block.BlockContinue
            try_continue(Parser.Block.ParserState state) {
            if (finished)
                return Parser.Block.BlockContinue.none();

            /* Blank line ends type 6 and type 7 blocks */
            if (state.is_blank() && closing_pattern == null) {
                return Parser.Block.BlockContinue.none();
            } else {
                return Parser.Block.BlockContinue.at_index(state.get_index());
            }
        }

        public override void add_line(Parser.SourceLine line) {
            content.add(line.content);

            if (closing_pattern != null && closing_pattern.match(line.content))
                finished = true;
        }

        public override void close_block() {
            block.literal = content.get_string();
            content = null;
        }

        public class Factory : Parser.Block.AbstractBlockParserFactory {

            public override Parser.Block.BlockStart try_start(
                Parser.Block.ParserState state,
                Parser.Block.MatchedBlockParser matched_block_parser) {
                int next_non_space = state.get_next_non_space_index();
                var line = state.get_line().content;

                if (state.get_indent() < 4 && line.get(next_non_space) == '<') {
                    for (int block_type = 1; block_type <= 7; block_type++) {
                        /* Type 7 can not interrupt a paragraph (not even a lazy one) */
                        if (block_type == 7 &&
                            (matched_block_parser.get_matched_block_parser()
                             is Node.Paragraph ||
                             state.get_active_block_parser()
                             .can_have_lazy_continuation_lines())) {
                            continue;
                        }
                        var opener = BLOCK_PATTERNS[block_type, 0];
                        var closer = BLOCK_PATTERNS[block_type, 1];
                        bool matches = opener.match(line.slice(next_non_space,
                                                               line.length));
                        if (matches) {
                            return Parser.Block.BlockStart.from(
                                new HtmlBlockParser(closer))
                                .at_index(state.get_index());
                        }
                    }
                }
                return Parser.Block.BlockStart.none();
            }
        }

        static construct {
            try {
                BLOCK_PATTERNS = {
                    { null, null}, /* not used (no type 0) */

                    { new GLib.Regex("^<(?:script|pre|style|textarea)(?:\\s|>|$)",
                                     GLib.RegexCompileFlags.CASELESS),
                      new GLib.Regex("</(?:script|pre|style|textarea)>",
                                     GLib.RegexCompileFlags.CASELESS) },

                    { new GLib.Regex("^<!--"),
                      new GLib.Regex("-->") },

                    { new GLib.Regex("^<[?]"),
                      new GLib.Regex("\\?>") },

                    { new GLib.Regex("^<![A-Z]"),
                      new GLib.Regex(">") },

                    { new GLib.Regex("^<!\\[CDATA\\["),
                      new GLib.Regex("\\]\\]>") },

                    { new GLib.Regex("^</?(?:" +
                                     "address|article|aside|" +
                                     "base|basefont|blockquote|body|" +
                                     "caption|center|col|colgroup|" +
                                     "dd|details|dialog|dir|div|dl|dt|" +
                                     "fieldset|figcaption|figure|footer|form|frame|frameset|" +
                                     "h1|h2|h3|h4|h5|h6|head|header|hr|html|" +
                                     "iframe|" +
                                     "legend|li|link|" +
                                     "main|menu|menuitem|" +
                                     "nav|noframes|" +
                                     "ol|optgroup|option|" +
                                     "p|param|" +
                                     "section|source|summary|" +
                                     "table|tbody|td|tfoot|th|thead|title|tr|track|" +
                                     "ul" +
                                     ")(?:\\s|[/]?[>]|$)",
                                     GLib.RegexCompileFlags.CASELESS),
                      null /* terminated by blank line */ },

                    { new GLib.Regex("^(?:" + Internal.Util.Parsing.OPEN_TAG + "|" +
                                     Internal.Util.Parsing.CLOSE_TAG + ")\\s*$",
                                     GLib.RegexCompileFlags.CASELESS),
                      null /* terminated by blank line */ }
                };
            } catch (GLib.Error error) {
                GLib.error(error.message);
            }
        }
    }
}
