

namespace Commonmark.Parser {

    /**
     * Parses input text to a tree of nodes.
     *
     * Start with the #builder method, configure the parser and build it. Example:
     * <pre><code>
     * Parser parser = Parser.builder().build();
     * Node document = parser.parse("input text");
     * </code></pre>
     */
    public class Parser : GLib.Object {

        private Gee.List<Block.BlockParserFactory> block_parser_factories;
        private Gee.List<Delimiter.DelimiterProcessor> delimiter_processors;
        private InlineParserFactory inline_parser_factory;
        private Gee.List<PostProcessor> post_processors;
        private IncludeSourceSpans include_source_spans;

        private Parser(Builder builder) {
            this.block_parser_factories =
                Internal.DocumentParser.calculate_block_parser_factories(
                    builder.block_parser_factories,
                    builder.enabled_block_types);
            this.inline_parser_factory = builder.get_inline_parser_factory();
            this.post_processors = builder.post_processors;
            this.delimiter_processors = builder.delimiter_processors;
            this.include_source_spans = builder.include_source_spans;

            /* Try to construct an inline parser. Invalid configuration might
               result in an exception, which we want to detect as soon as
               possible. */
            this.inline_parser_factory.create(
                new Internal.InlineParserContextImpl(
                    delimiter_processors,
                    new Internal.LinkReferenceDefinitions()));
        }

        /**
         * Create a new builder for configuring a Parser.
         */
        public static Builder builder() {
            return new Builder();
        }

        /**
         * Parse the specified input text into a tree of nodes.
         *
         * This method is thread-safe (a new parser state is used for each
         * invocation).
         */
        public Node.Node parse(string input) {
            if (input == null)
                GLib.error("input must not be null");
            var document_parser = create_document_parser();
            var document = document_parser.parse(input);
            return post_process(document);
        }

        /**
         * Parse the specified reader into a tree of nodes. The caller is
         * responsible for closing the reader.
         *
         * <pre><code>
         * Parser parser = Parser.builder().build();
         * try (InputStreamReader reader =
         *     new InputStreamReader(
         *         new FileInputStream("file.md"), StandardCharsets.UTF_8)) {
         *     Node document = parser.parseReader(reader);
         *     ...
         * }
         * </code></pre>
         *
         * Note that if you have a file with a byte order mark (BOM), you need
         * to skip it before handing the reader to this library. There's
         * existing classes that do that, e.g. see {@code BOMInputStream} in
         * Commons IO.
         *
         * This method is thread-safe (a new parser state is used for each
         * invocation).
         *
         * @param input the reader to parse - must not be null
         * @return the root node
         * @throws IOException when reading throws an exception
         */
        public Node.Node parse_reader(GLib.InputStream input)
            throws GLib.Error {
            if (input == null)
                GLib.error("input must not be null");

            var document_parser = create_document_parser();
            var document = document_parser.parse_input(input);
            return post_process(document);
        }

        private Commonmark.Internal.DocumentParser create_document_parser() {
            return new Commonmark.Internal.DocumentParser(
                block_parser_factories, inline_parser_factory,
                delimiter_processors, include_source_spans);
        }

        private Node.Node post_process(Node.Node document) {
            Node.Node d = document;
            foreach (var post_processor in post_processors) {
                d = post_processor.process(document);
            }
            return d;
        }

        /**
         * Builder for configuring a {@link Parser}.
         */
        public class Builder : GLib.Object {

            public Gee.List<Block.BlockParserFactory>
                block_parser_factories;
            public Gee.List<Delimiter.DelimiterProcessor>
                delimiter_processors;
            public Gee.List<PostProcessor> post_processors;
            public Gee.Set<GLib.Type> enabled_block_types;
            public InlineParserFactory inline_parser_factory;
            public IncludeSourceSpans include_source_spans;

            construct {
                block_parser_factories =
                    new Gee.ArrayList<Block.BlockParserFactory>();
                delimiter_processors =
                    new Gee.ArrayList<Delimiter.DelimiterProcessor>();
                post_processors = new Gee.ArrayList<PostProcessor>();
                enabled_block_types =
                    Internal.DocumentParser.get_default_block_parser_types();
                include_source_spans = IncludeSourceSpans.NONE;
            }

            /**
             * @return the configured Parser
             */
            public Parser build() {
                return new Parser(this);
            }

            /**
             * @param extensions extensions to use on this parser
             * @return this
             */
            public Builder extensions(Gee.Iterable<Extension> extensions) {
                if (extensions == null)
                    GLib.error("extensions must not be null");
                foreach (var extension in extensions) {
                    if (extension is ParserExtension) {
                        var parser_extension = extension as ParserExtension;
                        parser_extension.extend(this);
                    }
                }
                return this;
            }

            /**
             * Describe the list of markdown features the parser will recognize
             * and parse.
             *
             * By default, CommonMark will recognize and parse the following set
             * of "block" elements:
             *
             * <ul>
             * <li>{@link Heading} ({@code #})
             * <li>{@link HtmlBlock} ({@code <html></html>})
             * <li>{@link ThematicBreak} (Horizontal Rule) ({@code ---})
             * <li>{@link FencedCodeBlock} ({@code ```})
             * <li>{@link IndentedCodeBlock}
             * <li>{@link BlockQuote} ({@code >})
             * <li>{@link ListBlock} (Ordered / Unordered List) ({@code 1. / *})
             * </ul>
             *
             * To parse only a subset of the features listed above, pass a list
             * of each feature's associated {@link Block} class.
             *
             * E.g., to only parse headings and lists:
             * <pre>
             *     {@code
             *     Parser.builder().enabledBlockTypes(new HashSet<>(Arrays.asList(Heading.class, ListBlock.class)));
             *     }
             * </pre>
             *
             * @param enabledBlockTypes A list of block nodes the parser will
             * parse.  If this list is empty, the parser will not recognize any
             * CommonMark core features.
             * @return {@code this}
             */
            public Builder set_enabled_block_types(
                Gee.Set<GLib.Type> enabled_block_types) {
                if (enabled_block_types == null)
                    GLib.error("enabledBlockTypes must not be null");
                Internal.DocumentParser
                    .check_enabled_block_types(enabled_block_types);
                this.enabled_block_types = enabled_block_types;
                return this;
            }

            /**
             * Whether to calculate {@link org.commonmark.node.SourceSpan} for
             * {@link Node}.
             *
             * By default, source spans are disabled.
             *
             * @param includeSourceSpans which kind of source spans should be included
             * @return {@code this}
             * @since 0.16.0
             */
            public Builder get_include_source_spans(
                IncludeSourceSpans include_source_spans) {
                this.include_source_spans = include_source_spans;
                return this;
            }

            /**
             * Adds a custom block parser factory.
             *
             * Note that custom factories are applied <em>before</em> the
             * built-in factories. This is so that extensions can change how
             * some syntax is parsed that would otherwise be handled by built-in
             * factories.  "With great power comes great responsibility."
             *
             * @param blockParserFactory a block parser factory implementation
             * @return {@code this}
             */
            public Builder custom_block_parser_factory(
                Block.BlockParserFactory block_parser_factory) {
                if (block_parser_factory == null)
                    GLib.error("blockParserFactory must not be null");
                block_parser_factories.add(block_parser_factory);
                return this;
            }

            /**
             * Adds a custom delimiter processor.
             *
             * Note that multiple delimiter processors with the same characters
             * can be added, as long as they have a different minimum length. In
             * that case, the processor with the shortest matching length is
             * used. Adding more than one delimiter processor with the same
             * character and minimum length is invalid.
             *
             * @param delimiterProcessor a delimiter processor implementation
             * @return {@code this}
             */
            public Builder custom_delimiter_processor(
                Delimiter.DelimiterProcessor delimiter_processor) {
                if (delimiter_processor == null)
                    GLib.error("delimiterProcessor must not be null");
                delimiter_processors.add(delimiter_processor);
                return this;
            }

            public Builder post_processor(PostProcessor post_processor) {
                if (post_processor == null)
                    GLib.error("postProcessor must not be null");
                post_processors.add(post_processor);
                return this;
            }

            /**
             * Overrides the parser used for inline markdown processing.
             *
             * Provide an implementation of InlineParserFactory which provides a
             * custom inline parser to modify how the following are parsed: bold
             * (**) italic (*) strikethrough (~~) backtick quote (`) link
             * ([title](http://)) image (![alt](http://))
             *
             * Note that if this method is not called or the inline parser
             * factory is set to null, then the default implementation will be
             * used.
             *
             * @param inlineParserFactory an inline parser factory implementation
             * @return {@code this}
             */
            public Builder set_inline_parser_factory(
                InlineParserFactory inline_parser_factory) {
                this.inline_parser_factory = inline_parser_factory;
                return this;
            }

            public InlineParserFactory get_inline_parser_factory() {
                if (inline_parser_factory != null)
                    return inline_parser_factory;
                return new SimpleInlineParserFactory();
            }

            public class SimpleInlineParserFactory : GLib.Object,
                InlineParserFactory {

                public InlineParser
                    create(InlineParserContext inline_parser_context) {
                    return new Commonmark.Internal.InlineParserImpl(
                        inline_parser_context);
                }
            }
        }

        /**
         * Extension for Parser.
         */
        public interface ParserExtension : GLib.Object, Extension {
            public abstract void extend(Builder parserBuilder);
        }
    }
}
