namespace Commonmark.Renderer.Html {

    /**
     * The node renderer that renders all the core nodes (comes last in the
     * order of node renderers).
     */
    public class CoreHtmlNodeRenderer : Node.AbstractVisitor,
        Commonmark.Renderer.NodeRenderer {

        protected HtmlNodeRendererContext context;
        private HtmlWriter html;

        public CoreHtmlNodeRenderer(HtmlNodeRendererContext context) {
            this.context = context;
            this.html = context.get_writer();
        }

        public Gee.Set<GLib.Type> get_node_types() {
            var nt = new Gee.HashSet<GLib.Type>();
            nt.add_all_array(
                new GLib.Type[] {
                    typeof(Node.Document),
                    typeof(Node.Heading),
                    typeof(Node.Paragraph),
                    typeof(Node.BlockQuote),
                    typeof(Node.BulletList),
                    typeof(Node.FencedCodeBlock),
                    typeof(Node.HtmlBlock),
                    typeof(Node.ThematicBreak),
                    typeof(Node.IndentedCodeBlock),
                    typeof(Node.Link),
                    typeof(Node.ListItem),
                    typeof(Node.OrderedList),
                    typeof(Node.Image),
                    typeof(Node.Emphasis),
                    typeof(Node.StrongEmphasis),
                    typeof(Node.Text),
                    typeof(Node.Code),
                    typeof(Node.HtmlInline),
                    typeof(Node.SoftLineBreak),
                    typeof(Node.HardLineBreak) });
            return nt;
        }

        public void render(Node.Node node) {
            node.accept(this);
        }

        public override void visit(Node.Node node) {
            if (node is Node.BlockQuote)
                visit_block_quote(node as Node.BlockQuote);
            else if (node is Node.BulletList)
                visit_bullet_list(node as Node.BulletList);
            else if (node is Node.Code)
                visit_code(node as Node.Code);
            else if (node is Node.Document)
                visit_children(node as Node.Document);
            else if (node is Node.Emphasis)
                visit_emphasis(node as Node.Emphasis);
            else if (node is Node.FencedCodeBlock)
                visit_fenced_code_block(node as Node.FencedCodeBlock);
            else if (node is Node.HardLineBreak)
                visit_hard_line_break(node as Node.HardLineBreak);
            else if (node is Node.Heading)
                visit_heading(node as Node.Heading);
            else if (node is Node.HtmlBlock)
                visit_html_block(node as Node.HtmlBlock);
            else if (node is Node.HtmlInline)
                visit_html_inline(node as Node.HtmlInline);
            else if (node is Node.Image)
                visit_image(node as Node.Image);
            else if (node is Node.IndentedCodeBlock)
                visit_indented_code_block(node as Node.IndentedCodeBlock);
            else if (node is Node.Link)
                visit_link(node as Node.Link);
            else if (node is Node.ListItem)
                visit_list_item(node as Node.ListItem);
            else if (node is Node.OrderedList)
                visit_ordered_list(node as Node.OrderedList);
            else if (node is Node.Paragraph)
                visit_paragraph(node as Node.Paragraph);
            else if (node is Node.SoftLineBreak)
                visit_soft_line_break(node as Node.SoftLineBreak);
            else if (node is Node.StrongEmphasis)
                visit_strong_emphasis(node as Node.StrongEmphasis);
            else if (node is Node.Text)
                visit_text(node as Node.Text);
            else if (node is Node.ThematicBreak)
                visit_thematic_break(node as Node.ThematicBreak);
            else
                GLib.error("Invalid node type");
        }

        private void visit_block_quote(Node.BlockQuote block_quote) {
            html.line();
            html.tag_attrs("blockquote", get_attrs(block_quote, "blockquote"));
            html.line();
            visit_children(block_quote);
            html.line();
            html.tag("/blockquote");
            html.line();
        }

        private void visit_bullet_list(Node.BulletList bullet_list) {
            render_list_block(bullet_list, "ul",
                              get_attrs(bullet_list, "ul"));
        }

        private void visit_code(Node.Code code) {
            html.tag_attrs("code", get_attrs(code, "code"));
            html.text(code.literal);
            html.tag("/code");
        }

        private void visit_emphasis(Node.Emphasis emphasis) {
            html.tag_attrs("em", get_attrs(emphasis, "em"));
            visit_children(emphasis);
            html.tag("/em");
        }

        private void visit_fenced_code_block(
            Node.FencedCodeBlock fenced_code_block) {
            string literal = fenced_code_block.literal;
            var attributes = new Util.LinkedMap<string, string>();
            string info = fenced_code_block.info;
            if (info != null && info.length > 0) {
                int space = info.index_of(" ");
                string language;
                if (space == -1) {
                    language = info;
                } else {
                    language = info.substring(0, space);
                }
                attributes["class"] = "language-" + language;
            }
            render_code_block(literal, fenced_code_block, attributes);
        }

        private void visit_hard_line_break(
            Node.HardLineBreak hard_line_break) {
            html.tag_full("br", get_attrs(hard_line_break, "br"), true);
            html.line();
        }

        private void visit_heading(Node.Heading heading) {
            string htag = "h%d".printf(heading.level);
            html.line();
            html.tag_attrs(htag, get_attrs(heading, htag));
            visit_children(heading);
            html.tag("/" + htag);
            html.line();
        }

        private void visit_html_block(Node.HtmlBlock html_block) {
            html.line();
            if (context.should_escape_html()) {
                html.tag_attrs("p", get_attrs(html_block, "p"));
                html.text(html_block.literal);
                html.tag("/p");
            } else {
                html.raw(html_block.literal);
            }
            html.line();
        }

        private void visit_html_inline(Node.HtmlInline html_inline) {
            if (context.should_escape_html()) {
                html.text(html_inline.literal);
            } else {
                html.raw(html_inline.literal);
            }
        }

        private void visit_image(Node.Image image) {
            string url = image.destination;

            var alt_text_visitor = new AltTextVisitor();
            image.accept(alt_text_visitor);
            string alt_text = alt_text_visitor.get_alt_text();

            var attrs = new Util.LinkedMap<string, string>();
            if (context.should_sanitize_urls()) {
                url = context.url_sanitizer().sanitize_image_url(url);
            }

            attrs["src"] = context.encode_url(url);
            attrs["alt"] = alt_text;
            if (image.title != null) {
                attrs["title"] = image.title;
            }

            html.tag_full("img", get_attrs_with_default(image, "img", attrs),
                          true);
        }

        private void visit_indented_code_block(
            Node.IndentedCodeBlock indented_code_block) {
            render_code_block(indented_code_block.literal,
                              indented_code_block,
                              new Util.LinkedMap<string, string>());
        }

        private void visit_link(Node.Link link) {
            var attrs = new Util.LinkedMap<string, string>();
            string url = link.destination;

            if (context.should_sanitize_urls()) {
                url = context.url_sanitizer().sanitize_link_url(url);
                attrs["rel"] = "nofollow";
            }

            url = context.encode_url(url);
            attrs["href"] = url;
            if (link.title != null) {
                attrs["title"] = link.title;
            }
            html.tag_attrs("a", get_attrs_with_default(link, "a", attrs));
            visit_children(link);
            html.tag("/a");
        }

        private void visit_list_item(Node.ListItem list_item) {
            html.tag_attrs("li", get_attrs(list_item, "li"));
            visit_children(list_item);
            html.tag("/li");
            html.line();
        }

        private void visit_ordered_list(Node.OrderedList ordered_list) {
            int start = ordered_list.start_number;
            var attrs = new Util.LinkedMap<string, string>();
            if (start != 1) {
                attrs["start"] = "%d".printf(start);
            }
            render_list_block(ordered_list, "ol",
                              get_attrs_with_default(ordered_list, "ol",
                                                     attrs));
        }

        private void visit_paragraph(Node.Paragraph paragraph) {
            bool in_tight_list = is_in_tight_list(paragraph);
            if (!in_tight_list) {
                html.line();
                html.tag_attrs("p", get_attrs(paragraph, "p"));
            }
            visit_children(paragraph);
            if (!in_tight_list) {
                html.tag("/p");
                html.line();
            }
        }

        private void visit_soft_line_break(Node.SoftLineBreak softLineBreak) {
            html.raw(context.get_softbreak());
        }

        private void visit_strong_emphasis(
            Node.StrongEmphasis strong_emphasis) {
            html.tag_attrs("strong", get_attrs(strong_emphasis, "strong"));
            visit_children(strong_emphasis);
            html.tag("/strong");
        }

        private void visit_text(Node.Text text) {
            html.text(text.literal);
        }

        private void visit_thematic_break(Node.ThematicBreak thematicBreak) {
            html.line();
            html.tag_full("hr", get_attrs(thematicBreak, "hr"), true);
            html.line();
        }

        protected override void visit_children(Node.Node parent) {
            var node = parent.first_child;
            while (node != null) {
                context.render(node);
                node = node.next;
            }
        }

        private void render_code_block(string literal,
                                       Node.Node node,
                                       Gee.Map<string, string> attributes) {
            html.line();
            html.tag_attrs("pre", get_attrs(node, "pre"));
            html.tag_attrs("code", get_attrs_with_default(node, "code",
                                                          attributes));
            html.text(literal);
            html.tag("/code");
            html.tag("/pre");
            html.line();
        }

        private void render_list_block(Node.ListBlock list_block,
                                       string tag_name,
                                       Gee.Map<string, string> attributes) {
            html.line();
            html.tag_attrs(tag_name, attributes);
            html.line();
            visit_children(list_block);
            html.line();
            html.tag("/" + tag_name);
            html.line();
        }

        private bool is_in_tight_list(Node.Paragraph paragraph) {
            var parent = paragraph.parent;
            if (parent != null) {
                var gramps = parent.parent;
                if (gramps is Node.ListBlock) {
                    Node.ListBlock list = gramps as Node.ListBlock;
                    return list.tight;
                }
            }
            return false;
        }

        private Gee.Map<string, string> get_attrs(Node.Node node,
                                                  string tag_name) {
            return get_attrs_with_default(node, tag_name,
                                          new Util.LinkedMap<string, string>());
        }

        private Gee.Map<string, string>
            get_attrs_with_default(Node.Node node, string tag_name,
                                   Gee.Map<string, string> default_attributes) {
            return context.extend_attributes(node, tag_name, default_attributes);
        }

        private class AltTextVisitor : Node.AbstractVisitor {

            private StringBuilder sb;

            construct {
                sb = new StringBuilder();
            }

            public string get_alt_text() {
                return sb.str;
            }

            public override void visit(Node.Node node) {
                if (node is Node.Text)
                    sb.append(((Node.Text)node).literal);
                else if (node is Node.SoftLineBreak ||
                         node is Node.HardLineBreak)
                    sb.append_c('\n');
                else {
                    base.visit(node);
                }
            }
        }
    }
}
