namespace Commonmark.Renderer.Html {

    /**
     * Renders a tree of nodes to HTML.
     *
     * Start with the {@link #builder} method to configure the
     * renderer. Example:
     *
     * <pre><code>
     * HtmlRenderer renderer = HtmlRenderer.builder().escapeHtml(true).build();
     * renderer.render(node);
     * </code></pre>
     */
    public class HtmlRenderer : GLib.Object, Renderer {

        private string softbreak;
        private bool escape_html;
        private bool sanitize_urls;
        private UrlSanitizer url_sanitizer;
        private bool _percent_encode_urls;
        private Gee.List<AttributeProviderFactory> attribute_provider_factories;
        private Gee.List<HtmlNodeRendererFactory> node_renderer_factories;

        private HtmlRenderer(Builder builder) {
            this.softbreak = builder._softbreak;
            this.escape_html = builder.escape_html;
            this.sanitize_urls = builder._sanitize_urls;
            this._percent_encode_urls = builder._percent_encode_urls;
            this.url_sanitizer = builder._url_sanitizer;
            this.attribute_provider_factories =
                new Gee.ArrayList<AttributeProviderFactory>();
            this.attribute_provider_factories.add_all(
                    builder.attribute_provider_factories);
            this.node_renderer_factories =
                new Gee.ArrayList<HtmlNodeRendererFactory>();
            this.node_renderer_factories.add_all(
                builder.node_renderer_factories);
            /* Add as last. This means clients can override the rendering of
               core nodes if they want. */
            this.node_renderer_factories.add(new SimpleFactory());
        }

        /**
         * Create a new builder for configuring an {@link HtmlRenderer}.
         *
         * @return a builder
         */
        public static Builder builder() {
            return new Builder();
        }

        public void render_to(Node.Node node, GLib.OutputStream output) {
            if (node == null)
                GLib.error("node must not be null");
            var context = new RendererContext(this, new HtmlWriter(output));
            context.render(node);
        }

        public string render(Node.Node node) {
            if (node == null)
                GLib.error("node must not be null");
            try {
                var os = new GLib.MemoryOutputStream(null,
                                                     GLib.realloc,
                                                     GLib.free);
                render_to(node, os);
                os.close();
                var data = os.steal_data();
                data.length = (int)os.get_data_size();
                return (string)data;
            } catch (GLib.Error error) {
                GLib.error(error.message);
            }
        }

        /**
         * Builder for configuring an {@link HtmlRenderer}. See methods for
         * default configuration.
         */
        public class Builder : GLib.Object {

            public string _softbreak = "\n";
            public bool escape_html = false;
            public bool _sanitize_urls = false;
            public UrlSanitizer _url_sanitizer = new DefaultUrlSanitizer();
            public  bool _percent_encode_urls = false;
            public Gee.List<AttributeProviderFactory>
                attribute_provider_factories =
                new Gee.ArrayList<AttributeProviderFactory>();
            public Gee.List<HtmlNodeRendererFactory>
                node_renderer_factories =
                new Gee.ArrayList<HtmlNodeRendererFactory>();

            /**
             * @return the configured {@link HtmlRenderer}
             */
            public HtmlRenderer build() {
                return new HtmlRenderer(this);
            }

            /**
             * The HTML to use for rendering a softbreak, defaults to {@code
             * "\n"} (meaning the rendered result doesn't have a line break).
             *
             * Set it to {@code "<br>"} (or {@code "<br />"} to make them hard
             * breaks.
             *
             * Set it to {@code " "} to ignore line wrapping in the source.
             *
             * @param softbreak HTML for softbreak
             * @return {@code this}
             */
            public Builder softbreak(string _softbreak) {
                this._softbreak = _softbreak;
                return this;
            }

            /**
             * Whether {@link HtmlInline} and {@link HtmlBlock} should be
             * escaped, defaults to {@code false}.
             *
             * Note that {@link HtmlInline} is only a tag itself, not the text
             * between an opening tag and a closing tag. So markup in the text
             * will be parsed as normal and is not affected by this option.
             *
             * @param escapeHtml true for escaping, false for preserving raw HTML
             * @return {@code this}
             */
            public Builder escapeHtml(bool escape_html) {
                this.escape_html = escape_html;
                return this;
            }

            /**
             * Whether {@link Image} src and {@link Link} href should be
             * sanitized, defaults to {@code false}.
             *
             * @param sanitizeUrls true for sanitization, false for preserving raw attribute
             * @return {@code this}
             * @since 0.14.0
             */
            public Builder sanitize_urls(bool _sanitize_urls) {
                this._sanitize_urls = _sanitize_urls;
                return this;
            }

            /**
             * {@link UrlSanitizer} used to filter URL's if {@link
             * #sanitizeUrls} is true.
             *
             * @param urlSanitizer Filterer used to filter {@link Image} src and
             * {@link Link}.
             * @return {@code this}
             * @since 0.14.0
             */
            public Builder url_sanitizer(UrlSanitizer _url_sanitizer) {
                this._url_sanitizer = _url_sanitizer;
                return this;
            }

            /**
             * Whether URLs of link or images should be percent-encoded,
             * defaults to {@code false}.
             *
             * If enabled, the following is done:
             *
             * <ul>
             * <li>Existing percent-encoded parts are preserved (e.g. "%20" is
             * kept as "%20")</li>
             * <li>Reserved characters such as "/" are preserved, except for "["
             * and "]" (see encodeURI in JS)</li>
             * <li>Unreserved characters such as "a" are preserved</li>
             * <li>Other characters such umlauts are percent-encoded</li>
             * </ul>
             *
             * @param percentEncodeUrls true to percent-encode, false for
             * leaving as-is
             * @return {@code this}
             */
            public Builder percent_encode_urls(bool _percent_encode_urls) {
                this._percent_encode_urls = _percent_encode_urls;
                return this;
            }

            /**
             * Add a factory for an attribute provider for adding/changing HTML
             * attributes to the rendered tags.
             *
             * @param attributeProviderFactory the attribute provider factory to
             * add
             * @return {@code this}
             */
            public Builder attribute_provider_factory(
                AttributeProviderFactory _attribute_provider_factory) {
                if (_attribute_provider_factory == null)
                    GLib.error("attributeProviderFactory must not be null");
                this.attribute_provider_factories.add(
                    _attribute_provider_factory);
                return this;
            }

            /**
             * Add a factory for instantiating a node renderer (done when
             * rendering). This allows to override the rendering of node types
             * or define rendering for custom node types.
             *
             * If multiple node renderers for the same node type are created,
             * the one from the factory that was added first "wins". (This is
             * how the rendering for core node types can be overridden; the
             * default rendering comes last.)
             *
             * @param nodeRendererFactory the factory for creating a node
             * renderer
             * @return {@code this}
             */
            public Builder node_renderer_factory(
                HtmlNodeRendererFactory _node_renderer_factory) {
                if (_node_renderer_factory == null)
                    GLib.error("nodeRendererFactory must not be null");
                this.node_renderer_factories.add(_node_renderer_factory);
                return this;
            }

            /**
             * @param extensions extensions to use on this HTML renderer
             * @return {@code this}
             */
            public Builder extensions(Gee.Iterable<Extension> _extensions) {
                if (_extensions == null)
                    GLib.error("extensions must not be null");
                foreach (var extension in _extensions) {
                    if (extension is HtmlRendererExtension) {
                        var html_renderer_extension =
                            (HtmlRendererExtension)extension;
                        html_renderer_extension.extend(this);
                    }
                }
                return this;
            }
        }

        /**
         * Extension for {@link HtmlRenderer}.
         */
        public interface HtmlRendererExtension : GLib.Object, Extension {
            public abstract void extend(Builder renderer_builder);
        }

        private class RendererContext : GLib.Object, HtmlNodeRendererContext,
            AttributeProviderContext {

            private HtmlRenderer html_renderer;
            private HtmlWriter html_writer;
            private Gee.List<AttributeProvider> attribute_providers;
            private Internal.Renderer.NodeRendererMap node_renderer_map =
                new Internal.Renderer.NodeRendererMap();

            public RendererContext(HtmlRenderer html_renderer,
                                    HtmlWriter html_writer) {
                this.html_renderer = html_renderer;
                this.html_writer = html_writer;

                attribute_providers = new Gee.ArrayList<AttributeProvider>();
                foreach (var attribute_provider_factory
                         in html_renderer.attribute_provider_factories) {
                    attribute_providers.add(
                        attribute_provider_factory.create(this));
                }

                /* The first node renderer for a node type "wins". */
                for (int i = html_renderer.node_renderer_factories.size - 1;
                     i >= 0; i--) {
                    var node_renderer_factory =
                        html_renderer.node_renderer_factories[i];
                    var node_renderer = node_renderer_factory.create(this);
                    node_renderer_map.add(node_renderer);
                }
            }

            public bool should_escape_html() {
                return html_renderer.escape_html;
            }

            public bool should_sanitize_urls() {
                return html_renderer.sanitize_urls;
            }

            public UrlSanitizer url_sanitizer() {
                return html_renderer.url_sanitizer;
            }

            public string encode_url(string url) {
                if (html_renderer._percent_encode_urls) {
                    return Internal.Util.Escaping.percent_encode_url(url);
                } else {
                    return url;
                }
            }

            public Gee.Map<string, string>
                extend_attributes(Node.Node node, string tag_name,
                                  Gee.Map<string, string> attributes) {
                var attrs = new Util.LinkedMap<string, string>();
                foreach (var key in attributes.keys)
                    attrs[key] = attributes[key];
                set_custom_attributes(node, tag_name, attrs);
                return attrs;
            }

            public HtmlWriter get_writer() {
                return html_writer;
            }

            public string get_softbreak() {
                return html_renderer.softbreak;
            }

            public void render(Node.Node node) {
                node_renderer_map.render(node);
            }

            private void set_custom_attributes(Node.Node node,
                                               string tag_name,
                                               Gee.Map<string, string> attrs) {
                foreach (var attribute_provider in attribute_providers) {
                    attribute_provider.set_attributes(node, tag_name, attrs);
                }
            }
        }

        public class SimpleFactory : GLib.Object, HtmlNodeRendererFactory {

            public NodeRenderer create(HtmlNodeRendererContext context) {
                return new CoreHtmlNodeRenderer(context);
            }
        }
    }
}
